const log = require('../../util/log');
const dgram = require('dgram');

class TelloProcessor {
    initialize () {
	this.telloIpAddress = '192.168.10.1';	
	this.maximumNumberOfRetries = 2;

	log.info(`Initialize tello with IP address '${this.telloIpAddress}'`);

        this.client = dgram.createSocket('udp4');
        this.server = dgram.createSocket('udp4');

        this.client.bind({
            address: '0.0.0.0',
            port: 8889,
            exclusive: true
        });

	log.info('Tello control client created');

	this.response = null;
	this.state = {};
	this.currentCommand = null;
	this.currentStartDate = null;
	this.currentRetryCount = 0;
	this.sleepUntil = null

	this.client.on('message', (message, remote) => {
	    const messageString = message.toString();
	    //log.info("Response '" +  messageString + "' received.");
	    this.response = messageString;
        }).on("listening", () => {
	    log.info('UDP client listening', this.client.address());
	}).on("error", (err) => {
	    log.info('UDP client error', err);
	}).on("close", () => {
	    log.info('UDP client close', err);
	});
	
        // Tello State
	this.server.on('message', (message, remote) => {
            // Example: mid:-1;x:-100;y:-100;z:-100;mpry:0,0,0;pitch:0;roll:1;yaw:82;vgx:0;vgy:0;vgz:0;templ:72;temph:75;tof:10;h:0;bat:63;baro:-28.43;time:187;agx:16.00;agy:-32.00;agz:-1001.00;
            const readableMessage = message.toString();
	    //log.info(readableMessage);
            this.parseState(readableMessage.slice(0, -1));
        });

        this.server.bind(8890, '0.0.0.0');
	
	log.info('Tello state server created');
    }
    
    setIpAddress (ipAddress) {
	this.telloIpAddress = ipAddress;
	log.info(`Tello IP address set to '${this.telloIpAddress}'`);
    }

    request (command, util, timeoutInMs = 7000) {
	if (this.isSleeping()) {
	    // Still sleeping: make scratch call this function again
	    util.yield();
	} else {
	    // Sleeping is over: process command
	    
	    if (this.currentCommand === null) {
		// No command being execurted: send given command
		log.info("Send and wait: command='" + command + "'");
		this.currentCommand = command;
		this.currentStartDate = new Date();
		this.sendCommand(command);
	    }
	    
	    if (command != this.currentCommand) {
		// Another command being executed: stop waiting for current one.
		log.info("Cannot execute commands '" + command + "' and '" + this.currentCommand + "' at the same time.");
		this.currentCommand = null;
		this.currentRetryCount = 0;
		return;
	    }
	    
	    var response = this.getResponse(timeoutInMs);
	    if (response === null) {
		// No response received yet: make scratch call this function again.
		util.yield();
	    } else {
		// Response received: evaluate it.
		log.info("Response on command '" + this.currentCommand + "' (retry=" + this.currentRetryCount + "): '" + response + "'.");
		
		if (response == 'ok' || response == 'OK,drone will reboot in 3s')
		{   // Response 'ok' received: reset current command variables.
		    this.currentCommand = null;	    
		    this.currentRetryCount = 0;
		    return;
		}
		else
		{   // Response is not 'ok': sleep first and retry.
		    if (this.currentRetryCount >= this.maximumNumberOfRetries) {
			// Maximum number of reties reached: stop executing the command.
			log.info("Failed to execute '" + this.currentCommand + "'.");
			this.currentCommand = null;
			this.currentRetryCount = 0;
			return;
		    } else {
			// Maximum number of reties not reached yet: make scratch call this function again and ensure that the command is resend
			this.currentCommand = null; // Ensure that the command is resent.
			this.currentRetryCount++;
			// Start sleeping 1s before resending the command: if 'error Not Joystick' is the response, and we resend the command immediately,
			// Tello does not perform anything just sends an 'ok' response.
			this.startSleeping(1000);  
			util.yield(); // make scratch call this function again
		    }
		}
	    }
	}
    }

    sendCommand (command) {	
	this.response = null
	
	const message = Buffer.from(command);
	
	log.info("Send command to tello: '" + message + "'");
	
        this.client.send(message, 0, message.length, 8889, this.telloIpAddress, (errorMessage, bytes) => {
	    if (errorMessage)
	    {
		throw new Error(errorMessage);
	    }
        });
    }

    getResponse (timeoutInMs) {
	if (this.response === null) {
	    if ((Date.now() - this.currentStartDate.getTime()) > timeoutInMs) {
                 this.response  = 'error - no response';
	    }
	}
	return this.response;
    }

    getState () {
        return this.state;
    }

    parseState (stateString) {
	this.state = {};
	// Variables are separated by ';'
        for (const variable of stateString.split(";")) {
	    // Variable name and value are separated by ':'.
	    var splitvariable = variable.split(':');
	    if (splitvariable[0] == "mpry") {
		// Variable 'mpry=pith,rol,yaw' contains the pitch, rol, yaw relative to detected mission pad
		var splitMpry =  splitvariable[1].split(",");
		this.state["mpitch"] = splitMpry[0];
		this.state["mroll"] = splitMpry[1];
		this.state["myaw"] = splitMpry[2];
	    }
	    else
	    {
		this.state[splitvariable[0]] = splitvariable[1];
	    }
        }
    }
    
    startSleeping (milliseconds) {
	this.sleepUntil = new Date(Date.now() + milliseconds);
	log.info(`Start sleeping ${milliseconds}ms`);
    }
		  

    isSleeping () {
	if (this.sleepUntil === null)
	{   
	    return false;
	}
	else
	{
	    if (Date.now() > this.sleepUntil.getTime())
	    {
		log.info('Stop sleeping');
		this.sleepUntil = null;
		return false;
	    }
	    else
	    {
		return true;
	    }
	}
    }
}
module.exports = TelloProcessor;
